/*
 * Copyright (c) 1997,1998 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdlib.h>

#include <mpi.h>

#include <transpose_mpi.h>
#include <fftwnd_mpi.h>

/***************************** Plan Creation ****************************/

fftwnd_mpi_plan fftwnd_mpi_create_plan(MPI_Comm comm,
				       int rank, const int *n,
				       fftw_direction dir,
				       int flags)
{
    fftwnd_mpi_plan p;

    if (rank < 2)
	return 0;

    p = (fftwnd_mpi_plan) fftw_malloc(sizeof(fftwnd_mpi_plan_data));
    p->p_fft = 0;
    p->p_transpose = 0;
    p->p_transpose_inv = 0;

    p->p_fft = fftwnd_create_plan(rank, n, dir, flags | FFTW_IN_PLACE);
    if (!p->p_fft)
	fftwnd_mpi_destroy_plan(p);

    p->p_transpose = transpose_mpi_create_plan(n[0], n[1], comm);
    if (!p->p_transpose)
	fftwnd_mpi_destroy_plan(p);

    p->p_transpose_inv = transpose_mpi_create_plan(n[1], n[0], comm);
    if (!p->p_transpose_inv)
	fftwnd_mpi_destroy_plan(p);

    return p;
}

fftwnd_mpi_plan fftw2d_mpi_create_plan(MPI_Comm comm,
				       int nx, int ny,
				       fftw_direction dir, int flags)
{
    int n[2];

    n[0] = nx;
    n[1] = ny;

    return fftwnd_mpi_create_plan(comm, 2, n, dir, flags);
}

fftwnd_mpi_plan fftw3d_mpi_create_plan(MPI_Comm comm,
				       int nx, int ny, int nz,
				       fftw_direction dir, int flags)
{
    int n[3];

    n[0] = nx;
    n[1] = ny;
    n[2] = nz;

    return fftwnd_mpi_create_plan(comm, 3, n, dir, flags);
}

/********************** Plan Destruction ************************/

void fftwnd_mpi_destroy_plan(fftwnd_mpi_plan p)
{
    if (p) {
	if (p->p_fft)
	    fftwnd_destroy_plan(p->p_fft);
	if (p->p_transpose)
	    transpose_mpi_destroy_plan(p->p_transpose);
	if (p->p_transpose_inv)
	    transpose_mpi_destroy_plan(p->p_transpose_inv);
	fftw_free(p);
    }
}

/********************* Getting Local Size ***********************/

void fftwnd_mpi_local_sizes(fftwnd_mpi_plan p,
			    int *local_nx,
			    int *local_x_start,
			    int *local_ny_after_transpose,
			    int *local_y_start_after_transpose,
			    int *total_local_size)
{
    if (p) {
	transpose_mpi_get_local_size(p->p_transpose->nx,
				     p->p_transpose->my_pe,
				     p->p_transpose->n_pes,
				     local_nx,
				     local_x_start);
	transpose_mpi_get_local_size(p->p_transpose->ny,
				     p->p_transpose->my_pe,
				     p->p_transpose->n_pes,
				     local_ny_after_transpose,
				     local_y_start_after_transpose);
	*total_local_size =
	    transpose_mpi_get_local_storage_size(p->p_transpose->nx,
						 p->p_transpose->ny,
						 p->p_transpose->my_pe,
						 p->p_transpose->n_pes);

	*total_local_size *= p->p_fft->n_after[1];
    }
}

/******************** Computing the Transform *******************/

void fftwnd_mpi_aux(fftwnd_plan p, int cur_dim,
		    int n_fields,
		    FFTW_COMPLEX *data)
{
     int n_after = p->n_after[cur_dim], n = p->n[cur_dim];
     int i;

     if (cur_dim == p->rank - 2) {
          /* just do the last dimension directly: */
          if (n_fields == 1)
               fftw(p->plans[p->rank - 1], n,
                    data, 1, n_after,
                    p->work, 1, 0);
	  else
	       for (i = 0; i < n; ++i)
		    fftw(p->plans[p->rank - 1], n_fields,
			 data + i*n_after*n_fields, n_fields, 1,
			 p->work, 1, 0);
     }
     else if (cur_dim != p->rank - 1) {	 /* at least 2 dimensions to go */
          /* process the subsequent dimensions recursively, in hyperslabs,
             to get maximum locality: */
          for (i = 0; i < n; ++i)
	       fftwnd_mpi_aux(p, cur_dim + 1,
			      n_fields,
			      data + i * n_after * n_fields);
     }

     /* do the current dimension: */
     if (p->nbuffers == 0) {
          fftw(p->plans[cur_dim], n_after * n_fields,
               data, n_after * n_fields, 1,
               p->work, 1, 0);
     }
     else     /* using contiguous copy buffers: */
          fftw_buffered(p->plans[cur_dim], n_after * n_fields,
                        data, n_after * n_fields, 1,
                        p->work, p->nbuffers, p->work + n);
}

void fftwnd_mpi(fftwnd_mpi_plan p,
		int n_fields, FFTW_COMPLEX * local_data,
		fftwnd_mpi_output_order output_order)
{
     int fft_iter;
     int j, i;
     fftwnd_plan p_fft;
     fftw_plan *plans;
     int *n, *n_after;
     int local_nx, local_ny;
     FFTW_COMPLEX *work;
     int step, el_size;
     
     if (n_fields <= 0)
	  return;
     
     p_fft = p->p_fft;
     plans = p_fft->plans;
     n = p_fft->n;
     n_after = p_fft->n_after;
     work = p_fft->work;
     local_nx = p->p_transpose->local_nx;
     
     /* do the second dimension */
     if (p_fft->nbuffers == 0)
	  fftw(plans[1], n_after[1] * n_fields * local_nx,
	       local_data, n_after[1] * n_fields * local_nx, 1,
	       work, 1, 0);
     else
	  fftw_buffered(plans[1], n_after[1] * n_fields * local_nx,
			local_data, n_after[1] * n_fields * local_nx, 1,
			work, p_fft->nbuffers, work + n[1]);
     
     el_size = 2 * n_fields * n_after[1];
     
     if (p_fft->rank == 2) /* just compute the transpose */
	  transpose_mpi(p->p_transpose, el_size, (FFTW_REAL *) local_data);
     else {
	  /* overlap FFTs of the remaining dimensions with the transpose: */
	  
	  FFTW_REAL *send_buf = transpose_allocate_send_buf(p->p_transpose, 
							    el_size);
	  
	  for (step = 0; step < p->p_transpose->num_steps; ++step) {
	       int block_y_start, block_ny;
	       FFTW_COMPLEX *block_data;
	       int nxy;
	       
	       transpose_get_send_block(p->p_transpose, step, 
					&block_y_start, &block_ny);
	       block_data = local_data
		    + n_fields * n_after[1] * local_nx * block_y_start;
	       nxy = local_nx * block_ny;
	       
	       if (p_fft->rank == 3) {
		    if (n_fields == 1)
			 fftw(plans[2], nxy, block_data, 1, n[2], work, 1, 0);
		    else
			 for (i = 0; i < nxy; ++i)
			      fftw(plans[2], n_fields,
				   block_data + i * n_fields * n[2],
				   n_fields, 1,
				   work, 1, 0);
	       }
	       else
		    for (i = 0; i < nxy; ++i)
			 fftwnd_mpi_aux(p_fft, 2, n_fields, 
					block_data
					+ i * n_fields * n_after[1]);
	       
	       transpose_finish_exchange_step(p->p_transpose, step - 1);
	       
	       transpose_start_exchange_step(p->p_transpose, el_size, 
					     (FFTW_REAL *) local_data,
					     send_buf, step);
	  }
	  
	  transpose_finish_exchange_step(p->p_transpose, step - 1);
	  
	  if (send_buf)
	       transpose_free(send_buf);
     }
     
     local_ny = p->p_transpose->local_ny;
     
     /* do the first dimension (now transposed and local) */
     if (p_fft->nbuffers == 0)
	  fftw(plans[0], n_after[1] * n_fields * local_ny,
	       local_data, n_after[1] * n_fields * local_ny, 1,
	       work, 1, 0);
     else
	  fftw_buffered(plans[0], n_after[1] * n_fields * local_ny,
			local_data, n_after[1] * n_fields * local_ny, 1,
			work, p_fft->nbuffers, work + n[0]);
     
     /* transpose back, if desired: */
     if (output_order == FFTW_NORMAL_ORDER)
	  transpose_mpi(p->p_transpose_inv, el_size, (FFTW_REAL *) local_data);
}
