/*
 * Copyright (c) 1997,1998 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *
 */

#include <stdio.h>
#include <math.h>

#include <mpi.h>
#include <fftwnd_mpi.h>

void zero_arr(FFTW_COMPLEX *a, int n)
{
     int i;
     for (i = 0; i < n; ++i)
          a[i].re = a[i].im = 0.0;
}

#define TIME_FFT(fft,t) \
{ \
     double ts,te; \
     double total_t; \
     int iters = 1,iter; \
     do { \
	  MPI_Barrier(MPI_COMM_WORLD); \
          ts = MPI_Wtime(); \
          for (iter = 0; iter < iters; ++iter) fft; \
          te = MPI_Wtime(); \
          t = (total_t = (te - ts)) / iters; \
          iters *= 2; \
     } while (total_t < 2.0); \
}

double mflops(int N, double t)
{
     return (5.0 * N * log((double) N) / (t * 1.0e6));
}

#define MIN_RANK 2
#define MAX_RANK 5

typedef struct {
     int rank;
     int n[MAX_RANK];
} dims;

int prod_dims(dims d)
{
     int i, N = 1;

     for (i = 0; i < d.rank; ++i)
	  N *= d.n[i];
     return N;
}

char *dims_string(dims d)
{
     int i;
     static char s[200];
     
     if (d.rank > 0)
	  sprintf(s,"%d",d.n[0]);
     for (i = 1; i < d.rank; ++i)
	  sprintf(s + strlen(s),"x%d",d.n[i]);
     return s;
}

int main(int argc, char **argv)
{
     dims testdims[] = {
	  { 2, { 512, 512 } },
	  { 2, { 1024, 1024 } },
	  { 2, { 1500, 1500 } },
	  { 3, { 16, 16, 16 } },
	  { 3, { 32, 32, 32 } },
	  { 3, { 64, 64, 64 } },
	  { 3, { 80, 80, 80 } },
	  { 3, { 100, 100, 100 } },
	  { 3, { 128, 128, 128 } }
     };
     int ntests = sizeof(testdims) / sizeof(dims);
     int i, test;
     FFTW_COMPLEX *arr;
     int max_size = 0;
     fftwnd_plan plan;
     fftwnd_mpi_plan mpi_plan;
     double t, t_mpi;
     int my_pe;
     dims command_line_dims;
     
     MPI_Init(&argc,&argv);
     MPI_Comm_rank(MPI_COMM_WORLD,&my_pe);

     fftw_die_hook = fftw_mpi_die;
     
     if (argc >= 2)
          command_line_dims.rank = atoi(argv[1]);

     if (command_line_dims.rank >= MIN_RANK 
	 && argc >= 2 + command_line_dims.rank) {
          for (i = 0; i < command_line_dims.rank; ++i) {
               command_line_dims.n[i] = atoi(argv[2 + i]);
               if (command_line_dims.n[i] <= 0)
                    fftw_die("Illegal param: dimensions must be positive.\n");
          }
	  testdims[0] = command_line_dims;
	  ntests = 1;
     }     

     for (i = 0; i < ntests; ++i) {
	  int N = prod_dims(testdims[i]);
	  if (N > max_size)
	       max_size = N;
     }
     
     arr = (FFTW_COMPLEX *) fftw_malloc(max_size * sizeof(FFTW_COMPLEX));

     if (!arr) {
        if (my_pe == 0)
	     fprintf(stderr,
		     "Not enough memory!  At least %d bytes needed.\n",
		     max_size * sizeof(FFTW_COMPLEX));
        fftw_die("out of memory.\n");
     }
     
     zero_arr(arr,max_size);

     if (my_pe == 0)
	  printf("%13s%13s%13s%13s%13s%13s\n", 
		 "array size", "fftwnd",
		 "fftwnd_mpi", "speedup",
		 "transposed", "speedup-tr.");
     
     for (test = 0; test < ntests; ++test) {
	  if (my_pe == 0) {
	       printf("%13s",dims_string(testdims[test]));
	       fflush(stdout);
	  }
	  
	  plan = fftwnd_create_plan(testdims[test].rank, testdims[test].n,
				    FFTW_FORWARD,
				    FFTW_MEASURE | FFTW_IN_PLACE
				    | FFTW_USE_WISDOM);

	  mpi_plan = fftwnd_mpi_create_plan(MPI_COMM_WORLD,
					    testdims[test].rank,
					    testdims[test].n,
					    FFTW_FORWARD,
					    FFTW_MEASURE | FFTW_IN_PLACE
					    | FFTW_USE_WISDOM);
	  
	  TIME_FFT(fftwnd(plan, 1, arr, 1, 0, arr, 1, 0), t);

	  if (my_pe == 0) {
	       printf("%13g",mflops(prod_dims(testdims[test]),t));
	       fflush(stdout);
	  }

	  TIME_FFT(fftwnd_mpi(mpi_plan, 1, arr, FFTW_NORMAL_ORDER), t_mpi);
	  
	  if (my_pe == 0) {
	       printf("%13g",mflops(prod_dims(testdims[test]),t_mpi));
	       printf("%13g", t / t_mpi);
	       fflush(stdout);
	  }
	  
	  TIME_FFT(fftwnd_mpi(mpi_plan, 1, arr, FFTW_TRANSPOSED_ORDER), t_mpi);
	  
	  if (my_pe == 0) {
	       printf("%13g",mflops(prod_dims(testdims[test]),t_mpi));
	       printf("%13g\n", t / t_mpi);
	       fflush(stdout);
	  }
	  
	  fftwnd_destroy_plan(plan);
	  fftwnd_mpi_destroy_plan(mpi_plan);
     }
     
     fftw_free(arr);
     
     MPI_Finalize();

     return 0;
}
