/*
 * Copyright (c) 1997,1998 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <stdlib.h>
#include <string.h>

#include <mpi.h>

#include <transpose_mpi.h>

#include "sched.h"

static int transpose_mpi_get_block_size(int n, int n_pes)
{
     if (n_pes > n)
	  n_pes = n;

     return((n + n_pes - 1) / n_pes);
}

void transpose_mpi_get_local_size(int n, int my_pe, int n_pes,
				  int *local_n, int *local_start)
{
     int block_size;

     if (n_pes > n)
	  n_pes = n;

     block_size = transpose_mpi_get_block_size(n,n_pes);
     *local_start = my_pe * block_size;

     if (*local_start >= n)
	  *local_n = 0;
     else if (*local_start + block_size >= n)
	  *local_n = n - *local_start;
     else
	  *local_n = block_size;
}

int transpose_mpi_get_local_storage_size(int nx, int ny,
					 int my_pe, int n_pes)
{
     int local_nx, local_ny, local_x_start, local_y_start;

     transpose_mpi_get_local_size(nx,my_pe,n_pes,&local_nx,&local_x_start);
     transpose_mpi_get_local_size(ny,my_pe,n_pes,&local_ny,&local_y_start);

     if (local_nx*ny > nx*local_ny)
	  return (local_nx*ny);
     else
	  return (nx*local_ny);
}

void transpose_die(const char *error_string)
{
     int my_pe;

     MPI_Comm_rank(MPI_COMM_WORLD, &my_pe);

     fprintf(stderr, "fftw process %d: %s", my_pe, error_string);

     MPI_Abort(MPI_COMM_WORLD, 1);
}

transpose_mpi_plan transpose_mpi_create_plan(int nx, int ny, 
					     MPI_Comm transpose_comm)
{
     transpose_mpi_plan p;
     int my_pe, n_pes, pe;
     int x_block_size, local_nx, local_x_start;
     int y_block_size, local_ny, local_y_start;
     transpose_mpi_exchange *exchange = 0;
     int step, send_block_size = 0, recv_block_size = 0, num_steps = 0;
     int **sched, sched_npes, sched_sortpe, sched_sort_ascending;
     MPI_Comm comm;

     /* create a new "clone" communicator so that transpose
	communications do not interfere with caller communications. */
     MPI_Comm_dup(transpose_comm, &comm);

     MPI_Comm_rank(comm,&my_pe);
     MPI_Comm_size(comm,&n_pes);

     x_block_size = transpose_mpi_get_block_size(nx,n_pes);
     transpose_mpi_get_local_size(nx,my_pe,n_pes,&local_nx,&local_x_start);
     y_block_size = transpose_mpi_get_block_size(ny,n_pes);
     transpose_mpi_get_local_size(ny,my_pe,n_pes,&local_ny,&local_y_start);

     if (local_nx > 0 || local_ny > 0) {

     sched_npes = n_pes;
     sched_sortpe = -1;
     for (pe = 0; pe < n_pes; ++pe) {
	  int pe_nx, pe_x_start, pe_ny, pe_y_start;

	  transpose_mpi_get_local_size(nx,pe,n_pes,
				       &pe_nx,&pe_x_start);
	  transpose_mpi_get_local_size(ny,pe,n_pes,
				       &pe_ny,&pe_y_start);

	  if (pe_nx == 0 && pe_ny == 0) {
	       sched_npes = pe;
	       break;
	  }
	  else if (pe_nx * y_block_size != pe_ny * x_block_size
		   && pe_nx != 0 && pe_ny != 0) {
	       if (sched_sortpe != -1)
		    transpose_die("BUG: More than one PE needs sorting!\n");
	       sched_sortpe = pe;
	       sched_sort_ascending = 
		    pe_nx * y_block_size > pe_ny * x_block_size;
	  }
     }

     sched = make_comm_schedule(sched_npes);
     if (!sched) {
	  MPI_Comm_free(&comm);
	  return 0;
     }

     if (sched_sortpe != -1) {
	  sort_comm_schedule(sched,sched_npes,sched_sortpe);
	  if (!sched_sort_ascending)
	       invert_comm_schedule(sched,sched_npes);
     }

     send_block_size = local_nx * y_block_size;
     recv_block_size = local_ny * x_block_size;

     num_steps = sched_npes;

     exchange = (transpose_mpi_exchange *)
	  transpose_malloc(num_steps * sizeof(transpose_mpi_exchange));
     if (!exchange) {
	  free_comm_schedule(sched,sched_npes);
	  MPI_Comm_free(&comm);
	  return 0;
     }

     for (step = 0; step < sched_npes; ++step) {
	  int dest_pe;
	  int dest_nx, dest_x_start;
	  int dest_ny, dest_y_start;

	  exchange[step].dest_pe = dest_pe =
	       exchange[step].block_num = sched[my_pe][step];

	  if (exchange[step].block_num == -1)
	       transpose_die("BUG: schedule ended too early.\n");
	  
	  transpose_mpi_get_local_size(nx,dest_pe,n_pes,
				       &dest_nx,&dest_x_start);
	  transpose_mpi_get_local_size(ny,dest_pe,n_pes,
				       &dest_ny,&dest_y_start);
	       
	  exchange[step].send_size = local_nx * dest_ny;
	  exchange[step].recv_size = dest_nx * local_ny;
     }

     free_comm_schedule(sched,sched_npes);

     } /* if (local_nx > 0 || local_ny > 0) */
     
     p = (transpose_mpi_plan) 
	  transpose_malloc(sizeof(transpose_mpi_plan_struct));
     if (!p) {
          transpose_free(exchange);
	  MPI_Comm_free(&comm);
	  return 0;
     }

     p->comm = comm;
     p->nx = nx; p->ny = ny;
     p->local_nx = local_nx; p->local_ny = local_ny;

     p->my_pe = my_pe; p->n_pes = n_pes;

     p->exchange = exchange;
     p->send_block_size = send_block_size;
     p->recv_block_size = recv_block_size;
     p->num_steps = num_steps;

     /* Set the type constant for passing to the MPI routines;
	here, we assume that TRANSPOSE_EL_TYPE is one of the
	floating-point types. */
     if (sizeof(TRANSPOSE_EL_TYPE) == sizeof(double))
	  p->el_type = MPI_DOUBLE;
     else if (sizeof(TRANSPOSE_EL_TYPE) == sizeof(float))
	  p->el_type = MPI_FLOAT;
     else
	  transpose_die("Unknown TRANSPOSE_EL_TYPE!\n");

     return p;
}

void transpose_mpi_destroy_plan(transpose_mpi_plan p)
{
     if (p) {
	  if (p->exchange) transpose_free(p->exchange);
	  MPI_Comm_free(&p->comm);
	  transpose_free(p);
     }
}

static void local_transpose_copy(TRANSPOSE_EL_TYPE *src, 
				 TRANSPOSE_EL_TYPE *dest,
				 int el_size, int nx, int ny)
{
     int x, y;

     if (el_size == 1)
	  for (x = 0; x < nx; ++x)
	       for (y = 0; y < ny; ++y)
		    dest[y * nx + x] = src[x * ny + y];
     else if (el_size == 2)
	  for (x = 0; x < nx; ++x)
	       for (y = 0; y < ny; ++y) {
		    dest[y * (2 * nx) + 2*x]     = src[x * (2 * ny) + 2*y];
		    dest[y * (2 * nx) + 2*x + 1] = src[x * (2 * ny) + 2*y + 1];
	       }
     else
	  for (x = 0; x < nx; ++x)
	       for (y = 0; y < ny; ++y)
		    memcpy(&dest[y * (el_size*nx) + (el_size*x)],
			   &src[x * (el_size*ny) + (el_size*y)],
			   el_size * sizeof(TRANSPOSE_EL_TYPE));

}

TRANSPOSE_EL_TYPE *transpose_allocate_send_buf(transpose_mpi_plan p,
					       int el_size)
{
     TRANSPOSE_EL_TYPE *send_buf = 0;

     /* allocate the send buffer: */
     if (p->send_block_size > 0) {
          send_buf = (TRANSPOSE_EL_TYPE *)
               transpose_malloc(p->send_block_size * el_size
                                * sizeof(TRANSPOSE_EL_TYPE));
          if (!send_buf)
               transpose_die("Out of memory!\n");
     }

     return send_buf;
}

void transpose_mpi(transpose_mpi_plan p, int el_size,
		   TRANSPOSE_EL_TYPE *local_data)
{
     int my_pe, n_pes;
     int nx, ny, local_nx, local_ny;
     transpose_mpi_exchange *exchange;
     int send_block_size, num_steps, step;
     int recv_block_size;
     MPI_Comm comm;
     MPI_Datatype el_type;
     TRANSPOSE_EL_TYPE *send_buf;

     /* we copy everything into local variables for convenience */

     local_nx = p->local_nx;
     local_ny = p->local_ny;

     if (local_nx > 0 || local_ny > 0) {

     my_pe = p->my_pe;
     n_pes = p->n_pes;
     nx = p->nx;
     ny = p->ny;

     exchange = p->exchange;
     send_block_size = p->send_block_size;
     recv_block_size = p->recv_block_size;
     num_steps = p->num_steps;
     comm = p->comm;
     el_type = p->el_type;

     /* allocate the send buffer: */
     send_buf = transpose_allocate_send_buf(p,el_size);

     /* now, exchange the blocks between processors
        -- note that the ordering of the exchanges is critically important
           (see sched.c for more information) */

     for (step = 0; step < num_steps; ++step) {
	  int block = exchange[step].block_num;

	  /* first, copy to send buffer in transposed order: */
	  if (exchange[step].send_size > 0) {
	       local_transpose_copy(local_data 
				    + el_size*send_block_size*block,
				    send_buf,
				    el_size,
				    exchange[step].send_size / local_nx,
				    local_nx);
	  }

	  if (exchange[step].dest_pe != my_pe) {
               MPI_Status status;

	       if (exchange[step].send_size > 0
		   && exchange[step].recv_size > 0)
		    MPI_Sendrecv(send_buf,
				 exchange[step].send_size * el_size,
				 el_type,
				 exchange[step].dest_pe, 0,
				 
				 local_data + 
				 el_size*recv_block_size*block,
				 exchange[step].recv_size * el_size,
				 el_type,
				 exchange[step].dest_pe, MPI_ANY_TAG,
				 
				 comm,
				 &status);
	       else if (exchange[step].send_size > 0) 
		    MPI_Send(send_buf,
                             exchange[step].send_size * el_size,
                             el_type,
                             exchange[step].dest_pe, 0,
                             comm);
	       else if (exchange[step].recv_size > 0)
                    MPI_Recv(local_data +
                             el_size*recv_block_size*block,
                             exchange[step].recv_size * el_size,
                             el_type,
                             exchange[step].dest_pe, MPI_ANY_TAG,
                             comm,
                             &status);
	  }
	  else if (exchange[step].recv_size > 0)
	       memcpy(local_data +
		      el_size*recv_block_size*block,
		      send_buf,
		      exchange[step].recv_size * el_size *
		      sizeof(TRANSPOSE_EL_TYPE));
     }

     if (send_buf)
	  transpose_free(send_buf);

     } /* if (local_nx > 0 || local_ny > 0) */
}

/* non-blocking routines for overlapping communication and computation: */

#define USE_SYNCHRONOUS_SEND 1
#if USE_SYNCHRONOUS_SEND
#define ISEND MPI_Issend
#else
#define ISEND MPI_Isend
#endif

void transpose_get_send_block(transpose_mpi_plan p, int step,
			      int *block_y_start, int *block_ny)
{
     if (p->local_nx > 0) {
	  *block_y_start = 
	       p->send_block_size / p->local_nx * p->exchange[step].block_num;
	  *block_ny = p->exchange[step].send_size / p->local_nx;
     }
     else {
	  *block_y_start = 0;
	  *block_ny = 0;
     }
}

void transpose_start_exchange_step(transpose_mpi_plan p,
				   int el_size,
				   TRANSPOSE_EL_TYPE *local_data,
				   TRANSPOSE_EL_TYPE *send_buf,
				   int step)
{
     if (p->local_nx > 0 || p->local_ny > 0) {
	  transpose_mpi_exchange *exchange = p->exchange;
	  int block = exchange[step].block_num;
	  int send_block_size = p->send_block_size;
	  int recv_block_size = p->recv_block_size;
	  
          /* first, copy to send buffer in transposed order: */
          if (exchange[step].send_size > 0) {
               local_transpose_copy(local_data
                                    + el_size*send_block_size*block,
                                    send_buf,
                                    el_size,
                                    exchange[step].send_size / p->local_nx,
                                    p->local_nx);
          }
	  
          if (exchange[step].dest_pe != p->my_pe) {
	       /* Note that we impose an ordering on the sends and
		  receives (lower pe sends first) so that we won't
		  have deadlock if Isend & Irecv are blocking in some
		  MPI implementation: */
	       
#define DO_ISEND { \
	       if (exchange[step].send_size > 0) \
		    ISEND(send_buf, \
			      exchange[step].send_size * el_size, \
			      p->el_type, \
			      exchange[step].dest_pe, 0, \
			      p->comm, \
			      &p->request[0]); \
               else \
                    p->request[0] = MPI_REQUEST_NULL; \
	       }
	       
	       if (p->my_pe < exchange[step].dest_pe)
		    DO_ISEND;
	       
	       if (exchange[step].recv_size > 0)
		    MPI_Irecv(local_data +
			      el_size*recv_block_size*block,
			      exchange[step].recv_size * el_size,
			      p->el_type,
			      exchange[step].dest_pe, MPI_ANY_TAG,
			      p->comm,
			      &p->request[1]);
	       else
		    p->request[1] = MPI_REQUEST_NULL;
	       
	       if (p->my_pe > exchange[step].dest_pe)
		    DO_ISEND;
	  }
	  else if (exchange[step].recv_size > 0)
	       memcpy(local_data +
		      el_size*recv_block_size*block,
		      send_buf,
		      exchange[step].recv_size * el_size *
		      sizeof(TRANSPOSE_EL_TYPE));
     }
}

void transpose_finish_exchange_step(transpose_mpi_plan p, int step)
{
     if ((p->local_nx > 0 || p->local_ny > 0) && step >= 0
	 && p->exchange[step].dest_pe != p->my_pe) {
	  MPI_Status status[2];
	  
	  MPI_Waitall(2,p->request,status);
     }
}
